/*
 * fwho - display login records
 *
 * synopsis:
 *	fwho [+format] [-C] [-l lastlog ..] [-u utmp ..] [-w wtmp ..] [-U utmpx ..] [-W wtmpx ..]
 *
 * description:
 *	fwho displays lastlog(5), utmp(5), wtmp(5) format data and
 *	timestamps of those files and the time the user logged in with
 *	endian changing capabilities for the time data. The files may
 *	be compressed by gzip(1). Additionally fwho also displays
 *	utmpx, wtmpx data via getutxent(3).
 *
 * options:
 *	+format	specify strftime(3) format, default "%Y/%m/%d.%H:%M:%S".
 *	-C	specify changing endian for the time data.
 *	-l	specify lastlog format files.
 *	-u	specify utmp format files.
 *	-w	specify wtmp format files.
 *	-U	specify utmpx format files.
 *	-W	specify wtmpx format files.
 *
 * Copyright (c) 2009 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>	/* lstat, etc */
#include <time.h>	/* localtime, etc */
#if defined(sun)
#include <lastlog.h>
#endif
#include <utmp.h>	/* lastlog, utmp, wtmp, etc */
#if defined(sun) || defined(__APPLE__)
#include <utmpx.h>	/* getutxent, etc */
#endif
#include <sys/param.h>	/* MAXPATHLEN, etc */
#if defined(DISABLE_ZLIB)
typedef FILE *File;
#else
#include <zlib.h>	/* gzopen, etc */
typedef gzFile File;
#undef feof
#define fopen gzopen
#define fread(p,s,n,f) gzread(f,p,(s)*(n))
#define feof gzeof
#define fclose gzclose
#endif
#if defined(__APPLE__)
#define utmpxname(path) ((!strcmp(_PATH_UTMPX, path))?1:0)
#endif
#if defined(sun)
#define _PATH_LASTLOG	"/var/log/lastlog"
#define _PATH_UTMP	UTMP_FILE
#define _PATH_WTMP	WTMP_FILE
#define _PATH_UTMPX	UTMPX_FILE
#define _PATH_WTMPX	WTMPX_FILE
#else
#if !defined(__APPLE__)
#define _PATH_UTMPX	""
#endif
#define _PATH_WTMPX	""
#endif

/* not completely the same as BSD strlcpy */
#define strlcpy(d, s, z) do {\
  strncpy(d, s, z);\
  if ((z) > 0)\
    d[(z)-1] = '\0';\
} while (0)

/* swap for arbitrary type */
#define swap(a,b) do {\
  struct tmp_t { char t[sizeof(a)]; } tmp;\
  tmp = *(struct tmp_t *)&(a);\
  *(struct tmp_t *)&(a) = *(struct tmp_t *)&(b);\
  *(struct tmp_t *)&(b) = tmp;\
} while(0)

/* change endian for arbitrary type */
#define chend(p) do {\
  int i;\
  for (i=0; i<sizeof(*(p))/2; i++)\
    swap(((char *)p)[i], ((char *)p)[sizeof(*(p))-i-1] );\
} while (0)

static char ftm[256] = "%Y/%m/%d.%H:%M:%S";
static int sw_chend = 0;

int usage(int rc)
{
  puts("usage: fwho [+format] [-C] [-l lastlog ..] [-u utmp ..] [-w wtmp ..] [-U utmpx ..] [-W wtmpx ..]");
  return rc;
}

int fprint_ftimestamp(const char *path, FILE *ofp)
{
  struct stat sb;
  time_t *ti[] = { &sb.st_ctime, &sb.st_mtime, /*&sb.st_atime,*/ };
  size_t i;
  char buf[256];

  if (lstat(path, &sb)) return 1;
  fprintf(ofp, "%s: ", path);
  for (i=0; i<sizeof(ti)/sizeof(ti[0]); i++) {
    strftime(buf, sizeof(buf), ftm, localtime(ti[i]));
    fprintf(ofp, "%s", buf);
    fprintf(ofp, "%s", (i != sizeof(ti)/sizeof(ti[0])-1) ? " - " : "\n");
  }
  return 0;
}

int fprint_lastlog(const char *path, FILE *ofp)
{
  File ifp;
  size_t z, i = 0, c = 0;
  struct lastlog u;
  char buf[256], line[sizeof(u.ll_line)+1], host[sizeof(u.ll_host)+1];

  if (!strcmp(path, "-")) ifp = stdin;
  else if (!(ifp = fopen(path, "r"))) return 1;
  else fprint_ftimestamp(path, ofp);
  while ((z=fread(&u, sizeof(u), 1, ifp))) {
    if (u.ll_time) {
      if (sw_chend) chend(&u.ll_time);
      strftime(buf, sizeof(buf), ftm, localtime(&u.ll_time));
      strlcpy(line, u.ll_line, sizeof(line));
      strlcpy(host, u.ll_host, sizeof(host));
      fprintf(ofp, "%s\t%lu\t%-8s\t%-16s\n", buf, i, line, host);
      c++;
    }
    i++;
    if (feof(ifp)) break;
  }
  fprintf(ofp, "entry/record: %lu/%lu\n\n", c, i);
  if ((FILE *)ifp != stdin)
    fclose(ifp);
  return 0;
}
int fprint_utmp(const char *path, FILE *ofp)
{
#if defined(sun)
#define utmp futmp
#endif
  File ifp;
  size_t z, i = 0, c = 0;
  struct utmp u;
  char buf[256], line[sizeof(u.ut_line)+1], name[sizeof(u.ut_name)+1];
#if defined(sun)
#else
  char host[sizeof(u.ut_host)+1];
#endif

  if (!strcmp(path, "-")) ifp = stdin;
  else if (!(ifp = fopen(path, "r"))) return 1;
  else fprint_ftimestamp(path, ofp);
  while ((z=fread(&u, sizeof(u), 1, ifp))) {
    if (u.ut_time) {
      if (sw_chend) chend(&u.ut_time);
      strftime(buf, sizeof(buf), ftm, localtime((time_t *)&u.ut_time));
      strlcpy(line, u.ut_line, sizeof(line));
      strlcpy(name, u.ut_name, sizeof(name));
#if defined(sun)
      fprintf(ofp, "%s\t%-8s\t%-8s\t\n", buf, line, name);
#else
      strlcpy(host, u.ut_host, sizeof(host));
      fprintf(ofp, "%s\t%-8s\t%-8s\t%-16s\n", buf, line, name, host);
#endif
      c++;
    }
    i++;
    if (feof(ifp)) break;
  }
  fprintf(ofp, "entry/record: %lu/%lu\n\n", c, i);
  if ((FILE *)ifp != stdin)
    fclose(ifp);
  return 0;
#undef utmp
}
int fprint_wtmp(const char *path, FILE *ofp)
{
  return fprint_utmp(path, ofp);
}
int fprint_utmpx(const char *path, FILE *ofp)
{
#if defined(sun) || defined(__APPLE__)
  size_t c = 0;
  struct utmpx *p;
  char buf[256], user[sizeof(p->ut_user)+1], line[sizeof(p->ut_line)+1];
#if defined(sun)
#else
  char host[sizeof(p->ut_host)+1];
#endif

  if (utmpxname(path) != 1) return 1;
  else fprint_ftimestamp(path, ofp);
  while ((p = getutxent())) {
    strftime(buf, sizeof(buf), ftm, localtime(&p->ut_tv.tv_sec));
    strlcpy(user, p->ut_user, sizeof(user));
    strlcpy(line, p->ut_line, sizeof(line));
#if defined(sun)
    fprintf(ofp, "%s\t%-8s\t%-8s\n", buf,
	    p->ut_user, p->ut_line);
#else
    strlcpy(host, p->ut_host, sizeof(host));
    fprintf(ofp, "%s\t%-8s\t%-8s\t%-16s\n", buf,
	    p->ut_user, p->ut_line, p->ut_host);
#endif
    c++;
  }
  endutxent();
  fprintf(ofp, "entry: %lu\n\n", c);
#endif
  return 0;
}
int fprint_wtmpx(const char *path, FILE *ofp)
{
  return fprint_utmpx(path, ofp);
}

int main(int argc, char *argv[])
{
  int a, i = 0;

  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--help")==0) return usage(0);
    else if (argv[a][0] == '+') strlcpy(ftm, &argv[a][1], sizeof(ftm));
    else if (!strcmp(argv[a], "-C")) sw_chend = !0;
    else if (!strcmp(argv[a], "-l"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_lastlog(argv[i], stdout);
    else if (!strcmp(argv[a], "-u"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_utmp(argv[i], stdout);
    else if (!strcmp(argv[a], "-w"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_wtmp(argv[i], stdout);
    else if (!strcmp(argv[a], "-U"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_utmpx(argv[i], stdout);
    else if (!strcmp(argv[a], "-W"))
      for (i=a+1; i<argc && (!strcmp(argv[i], "-") || argv[i][0] != '-'); i++)
	fprint_wtmpx(argv[i], stdout);

  if (!i) {
    char wtmp[][MAXPATHLEN] = {
      _PATH_WTMP".6.gz",
      _PATH_WTMP".5.gz",
      _PATH_WTMP".4.gz",
      _PATH_WTMP".3.gz",
      _PATH_WTMP".2.gz",
      _PATH_WTMP".1.gz",
      _PATH_WTMP".0.gz",
      _PATH_WTMP,
    };
    char wtmpx[][MAXPATHLEN] = {
      _PATH_WTMPX,
    };

    fprint_lastlog(_PATH_LASTLOG, stdout);
    fprint_utmp(_PATH_UTMP, stdout);
    for (i=0; i<sizeof(wtmp)/sizeof(wtmp[0]); i++)
      fprint_wtmp(wtmp[i], stdout);
    fprint_utmpx(_PATH_UTMPX, stdout);
    for (i=0; i<sizeof(wtmpx)/sizeof(wtmpx[0]); i++)
      fprint_wtmpx(wtmpx[i], stdout);
  }
  return 0;
}
